# Start the imports
import os
import wx
# This is needed to create a temporary file for the compilation process
import tempfile
# I ue this only on Unix/Linux/Mac... but why Windows doesn't have os.kill?
import signal

from Utilities import opj


class Process(object):
    
    def __init__(self, parent, buildDir, setupScript, run, compiler,
                 project, pageNumber):
        """
        Default class constructor.

        @param: parent: the parent widget (GUI2Exe)
        @param buildDir: the directory in which we build the executable
        @param setupScript: the setup file passed as a string
        @param run: if False, only a "dry-run" is performed
        @param compiler: the compiler used (at the moment only py2exe)
        @param project: the project we are compiling
        @param pageNumber: the main wx.aui.AuiNotebook page number we belong to

        """        

        # Store the input data passed in __init__
        self.MainFrame = parent
        self.buildDir = buildDir
        self.setupScript = setupScript
        self.run = run
        self.compiler = compiler
        self.project = project
        self.pageNumber = pageNumber

        # No process started, the main frame will take care of it        
        self.process = None
        # Let's keep track if we failed the compilation step or not
        self.failed = False
        # We store all the output that comes from the process
        self.fullText = ""
        

    def Start(self):
        """ Starts the compilation process. """

        # Change the directory to the compilation folder
        os.chdir(opj(self.buildDir))
        # Verify if it is a dry-run or a real compilation
        dryRun = (self.run and [""] or ["--dry-run"])[0]

        # Create a temporary file that we will delete later        
        fd, tmpFileName = tempfile.mkstemp(suffix=".py", dir=self.buildDir)
        # Write the setup script string in the temporary file
        fid = open(tmpFileName, "wt")
        fid.write(self.setupScript)
        fid.close()
        # Store the temporary file data
        self.tmpFileName = (fd, tmpFileName)

        # Build the compilation command        
        cmd = 'python -u "%s" %s %s'%(tmpFileName, self.compiler, dryRun)

        # Start the process, redirecting the output to catch stdio
        self.process = wx.Process(self.MainFrame)
        self.process.Redirect()
        # We want the process to be asynchronous
        self.pid = wx.Execute(cmd, wx.EXEC_ASYNC, self.process)


    def Kill(self):
        """ Kills the process. """

        # This doesn't work very well on Windows :-(        
        self.process.CloseOutput()
        self.process.Kill(self.pid, wx.SIGTERM)
        if wx.Platform in ["__WXGTK__", "__WXMAC__"]:
            try:
                os.kill(self.pid, signal.SIGTERM)
            except:
                pass
            
        self.Destroy()
        

    def Destroy(self):
        """ Destroy the process. """

        # Ok, this line is commented because if I leave it as it is Python
        # bombs with the classical Windows error dialog
##        self.process.Destroy()
        
        self.MainFrame.process = None
        # Close the temporary file
        os.close(self.tmpFileName[0])
        # Remove the temporary file from the compilation directory
        if os.path.isfile(opj(self.tmpFileName[1])):
            try:
                os.remove(opj(self.tmpFileName[1]))
            except:
                # Something is keeping it open... skip for now
                pass

        # Returns to the installation directory            
        os.chdir(self.MainFrame.installDir)

        
    def HandleProcessMessages(self, processEnded=False):
        """ Handles all the messages that come from input and error streams. """

        # Get the input stream
        istream = self.process.GetInputStream()
        self.ReadStream(istream, 0, processEnded)
        # Get the error stream
        estream = self.process.GetErrorStream()
        self.ReadStream(estream, 1, processEnded)

        if processEnded and not self.failed:
            # Process finished without errors, send a congratulation message
            self.MainFrame.SendMessage("Message", "Setup file succesfully compiled")
            # Update visually the project page
            self.MainFrame.UpdatePageBitmap(self.project.GetName(), 1, self.pageNumber)
            # Process the output text from the compilation steps
            self.ProcessOutputText()
            if self.run:
                # It was a real compilation, ask the user to test the executable
                self.MainFrame.SuccessfulCompilation(self.project)


    def ReadStream(self, stream, isError, processEnded):
        """ Reads the input/error streams (if any). """
        
        if stream.CanRead():
            # There is some data, proces it
            text = stream.read()
            if isError:
                # Ah, is the error stream, something went wrong
                self.MainFrame.SendMessage("Error", text)
                self.failed = True
                self.fullText = ""
            else:
                # That's the input stream
                if text.find("byte-compiling") >= 0:    # py2exe is compiling
                    self.MainFrame.SendMessage("Compile", "Byte-compiling Python files...")
                elif text.find("copying") >= 0:         # py2exe is copying files
                    self.MainFrame.SendMessage("Copy", "Copying files...")
                elif text.find("searching") >= 0:       # py2exe is searching for modules
                    self.MainFrame.SendMessage("Find", "Finding required modules...")

                # Store the text from the input stream
                self.fullText += text


    def ProcessOutputText(self):
        """ Process the result of the compilation steps. """

        # Look for what py2exe thinks are the missing modules
        indx = self.fullText.rfind("The following modules")
        text = self.fullText[indx:]
        # They are usually put inside square braces
        indx1, indx2 = text.find("["), text.find("]")
        if indx1 < 0 or indx2 < 0:
            missingModules = []
        else:
            missingModules = eval(text[indx1:indx2+1])
            missingModules = [tuple(miss.split(".")) for miss in missingModules]

        # Look for what py2exe thinks are the binary dependencies
        binaryDependencies = []
        indx = text.find("*** binary dependencies ***")
        text = text[indx:].split("\n")
        text = text[7:]
        
        # Here it's bit tricky, but t works
        for line in text:
            line = line.strip().split("-")
            dll, path = line[0], "".join(line[1:])
            if not dll:
                break
            binaryDependencies.append([dll.strip(), opj(path.strip())])

        self.project.hasBeenCompiled = True
        self.project.SetCompilationData(missingModules, binaryDependencies)


