# Start the imports
import os
import sys
import wx
import wx.lib.mixins.listctrl as listmix
import wx.stc as stc
import wx.lib.buttons as buttons

# This is needed by the StyledTextCtrl
import keyword

from Utilities import opj, flatten, unique
from Constants import StringType, ListType, _iconFromName, _unWantedLists, _faces

                 
class BaseListCtrl(wx.ListCtrl, listmix.ListCtrlAutoWidthMixin, listmix.TextEditMixin,
                   listmix.ColumnSorterMixin):

    def __init__(self, parent, columnNames, name="", mainFrame=None):
        """
        Default class constructor.

        @param parent: the parent widget
        @param columnNames: the list control column names
        @param name: the list control name
        @param mainFrame: the application main frame (GUI2Exe)

        """

        wx.ListCtrl.__init__(self, parent, style=wx.LC_REPORT|wx.LC_HRULES|wx.LC_VRULES|wx.SUNKEN_BORDER,
                             name=name)

        # Initialize the auto width mixin. We always need it        
        listmix.ListCtrlAutoWidthMixin.__init__(self)
        if name not in _unWantedLists:
            # But we don't always want text edit mixin
            listmix.TextEditMixin.__init__(self)

        # If we are on a secondary frame, we pass the main frame reference
        if not mainFrame:
            # That's  secondary frame
            self.MainFrame = wx.GetTopLevelParent(self)
        else:
            # We are a descendant of GUI2Exe main frame
            self.MainFrame = mainFrame

        # The column sorter mixin will be initialized later
        self.columnSorter = False
        self.itemDataMap = {}

        # Do the hard work        
        self.BuildImageList()
        self.InsertColumns(columnNames)
        self.BindEvents()

        # Usually the "resources" lists for py2exe do not contain many items
        # so we allow them to be smaller than other lists
        if name.find("resources") < 0:
            self.SetMinSize((-1, 150))        


    # ========================== #
    # Methods called in __init__ #
    # ========================== #
    
    def InsertColumns(self, columnNames):
        """ Inserts the columns in the list control. """

        # The first column is always empty text, as I use it to display
        # an informative/fancy icon
        self.InsertColumn(0, "")

        # Loop over all the column names        
        for indx, column in enumerate(columnNames):
            self.InsertColumn(indx+1, column)
            self.SetColumnWidth(indx+1, wx.LIST_AUTOSIZE_USEHEADER)

        # The first column only displays an icon, 24 is perfect on Windows
        self.SetColumnWidth(0, 24)


    def BuildImageList(self):
        """ Build the image list for the list control. """

        # Ok, here it gets tricky as I am trying to re-use the same base class
        # for 6 or more different list control
        name = self.GetName()
        imgList = wx.ImageList(16, 16)
        # Get the icons from the list control name
        pngs = _iconFromName[name]

        # Loop over all the images in the list
        for png in pngs:
            imgList.Add(self.MainFrame.CreateBitmap(png))

        if name != "messages":
            # The other list can be sorted, but not the one at the bottom
            # that is our log message window
            self.sm_up = imgList.Add(self.MainFrame.CreateBitmap("sort_up"))
            self.sm_dn = imgList.Add(self.MainFrame.CreateBitmap("sort_down"))

        # Assign the image list, I don't want to store it            
        self.AssignImageList(imgList, wx.IMAGE_LIST_SMALL)            
        

    def BindEvents(self):
        """ Binds the event handler for the list. """

        self.Bind(wx.EVT_LIST_COL_BEGIN_DRAG, self.OnColumnDrag)
        
        if self.GetName() not in _unWantedLists:
            # For most of the list we really do hard work
            self.Bind(wx.EVT_LIST_BEGIN_LABEL_EDIT, self.OnBeginEdit)
            self.Bind(wx.EVT_LIST_END_LABEL_EDIT, self.OnEndEdit)
            self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
            self.Bind(wx.EVT_LIST_COL_RIGHT_CLICK, self.OnRightClick)
            self.Bind(wx.EVT_LIST_ITEM_RIGHT_CLICK, self.OnRightClick)
            self.popupId1, self.popupId2 = wx.NewId(), wx.NewId()
            self.Bind(wx.EVT_MENU, self.OnDeleteSelected, id=self.popupId1)
            self.Bind(wx.EVT_MENU, self.OnClearAll, id=self.popupId2)


    # ============== #
    # Event handlers #
    # ============== #
    
    def OnColumnDrag(self, event):
        """ Handles the wx.EVT_LIST_COL_BEGIN_DRAG event for the list control. """

        if event.GetColumn() == 0:
            # Veto the event for the first column, it holds the icon
            event.Veto()
            return

        event.Skip()


    def OnBeginEdit(self, event):
        """ Handles the wx.EVT_LIST_BEGIN_LABEL_EDIT event for the list control. """

        if event.GetColumn() == 0:
            # Veto the event for the first column, it holds the icon
            event.Veto()
            return

        if self.GetName() == "other_resources":
            if event.GetItem().GetId() == 0 and self.HasManifest():
                # You can't edit the manifest item (if present)
                event.Veto()
                return
            
        event.Skip()


    def OnEndEdit(self, event):
        """ Handles the wx.EVT_LIST_END_LABEL_EDIT event for the list control. """
        
        if event.IsEditCancelled():
            # Nothing to do, the user cancelled the editing
            event.Skip()
            return

        event.Skip()

        # Update the project, as something changed
        self.UpdateProject()

        # Adjust the data for the column sorter mixin
        indx = event.GetItem().GetId()
        tuple = ()
        # Loop over all the columns, populating a tuple
        for col in xrange(self.GetColumnCount()):
            item = self.GetItem(indx, col)
            tuple += (item.GetText(),)

        # Store the data
        self.SetItemData(indx, indx)
        self.itemDataMap[indx] = tuple
        # Initialize the column sorter mixin (if needed)
        self.InizializeSorter()


    def OnKeyDown(self, event):
        """ Handles the wx.EVT_KEY_DOWN event for the list control. """

        if event.GetModifiers() == wx.MOD_CONTROL and event.GetKeyCode() == ord("A"):
            # User pressed Ctrl+A, so he/she can add items to the list
            # The behavior is different depending on the list name
            name = self.GetName()
            if name.find("resources") >= 0:
                # New resource files, use a file dialog to pick them
                self.HandleNewResource(name)
            elif name == "data_files":
                # New data files, use a file dialog to pick them
                self.HandleDataFiles()
            else:
                # New modules, simply enter one item and start editing
                self.HandleNewModule()

        elif event.GetKeyCode() == wx.WXK_DELETE:
            # We are deleting something from the list control
            selections = self.GetSelectedIndices()
            # Reverse them, to delete them safely
            selections.reverse()
            # Loop over all the indices
            self.Freeze()
            for ind in selections:
                # Pop the data from the column sorter mixin dictionary
                indx = self.GetItemData(ind)
                self.itemDataMap.pop(indx)
                # Delete the item from the list
                self.DeleteItem(ind)
            self.Thaw()
            self.UpdateProject()
        # Otherwise skip the event
        event.Skip()


    def OnRightClick(self, event):
        """ Handles the wx.EVT_LIST_ITEM_RIGHT_CLICK event for the list control. """
        
        menu = wx.Menu()
        # Well, we can either delete the selected item(s)...
        item = wx.MenuItem(menu, self.popupId1, "Delete selected")
        bmp = self.MainFrame.CreateBitmap("delete_selected")
        item.SetBitmap(bmp)
        menu.AppendItem(item)
        # Or clear completely the list
        item = wx.MenuItem(menu, self.popupId2, "Clear all")
        bmp = self.MainFrame.CreateBitmap("clear_all")
        item.SetBitmap(bmp)
        menu.AppendItem(item)

        # Popup the menu on ourselves
        self.PopupMenu(menu)
        menu.Destroy()


    def OnDeleteSelected(self, event):
        """ Handles the wx.EVT_MENU event for the list control. """

        # Freeze everything... it helps with flicker
        self.Freeze()
        wx.BeginBusyCursor()
        # Get all the selected items
        indices = self.GetSelectedIndices()
        # Reverse them, to delete them safely
        indices.reverse()

        # Loop over all the indices        
        for ind in indices:
            # Pop the data from the column sorter mixin dictionary
            indx = self.GetItemData(ind)
            self.itemDataMap.pop(indx)
            # Delete the item from the list
            self.DeleteItem(ind)

        # Update the project, something changed
        self.UpdateProject()
        # Time to warm up...
        self.Thaw()
        wx.EndBusyCursor()
        

    def OnClearAll(self, event):
        """ Handles the wx.EVT_MENU event for the list control. """

        wx.BeginBusyCursor()
        # Freeze everything... it helps with flicker
        self.Freeze()
        # Clear all the list
        self.DeleteAllItems()
        # Time to warm up...
        self.Thaw()

        # Update the project, something changed
        self.itemDataMap = {}
        self.UpdateProject()
        wx.EndBusyCursor()
        

    # ================= #
    # Auxiliary methods #
    # ================= #
    
    def PopulateList(self, configuration):
        """ Populates the list control based on the input configuration. """

        if not configuration or not configuration[0]:
            # Nothing in here, go back
            return

        colCount = self.GetColumnCount()
        # This gets very tricky        
        for value in configuration:
            # The first item is always an empty string with the informative icon
            indx = self.InsertImageStringItem(sys.maxint, "", 0)
            # Initialize the tuple for the column sorter mixin
            tupleMap = ("",)
            if type(value) == StringType:
                # So, this must be one of "includes", "excludes", "modules",
                # "packages", "ignores", they are a list of strings
                self.SetStringItem(indx, 1, value.strip())
                tupleMap += (value.strip(),)
                self.itemDataMap[indx] = tupleMap
                self.SetItemData(indx, indx)
            else:
                # This is a list of lists
                if self.GetName() == "data_files":
                    # We have data files, a bit trickier
                    # Everything here is done to comply with the format in which
                    # the data is stored in the database
                    for index, tup in enumerate(value[1]):
                        if index > 0:
                            indx = self.InsertImageStringItem(sys.maxint, "", 0)
                            tupleMap = ("",)
                        self.SetStringItem(indx, 1, value[0])
                        self.SetStringItem(indx, 2, tup)
                        tupleMap += (value[0], tup)
                        self.itemDataMap[indx] = tupleMap
                        self.SetItemData(indx, indx)
                        
                else:
                    # Otherwise is a "resources" list
                    for col in xrange(colCount):
                        item = (col < len(value) and [value[col]] or [""])[0]
                        item = ("%s"%item).strip()
                        self.SetStringItem(indx, col+1, item)
                        tupleMap += (item,)

                    self.itemDataMap[indx] = tupleMap
                    self.SetItemData(indx, indx)

        # Initialize the column sorter mixin (if needed)
        self.InizializeSorter()


    def InizializeSorter(self):
        """
        Initializes the column sorter mixin.
        Not used by the "message" list.
        """

        # We don't want to sort our log window at the bottom        
        if not self.columnSorter and self.GetName() != "messages":
            # Not initialized yet
            colCount = self.GetColumnCount()
            listmix.ColumnSorterMixin.__init__(self, colCount)
            sortedColumn = (wx.GetTopLevelParent(self)==self.MainFrame and [colCount-1] or [1])[0]
            self.SortListItems(sortedColumn, True)
            self.columnSorter = True
            

    def GetListCtrl(self):
        """
        Used by the ColumnSorterMixin, see wx/lib/mixins/listctrl.py.
        Not used by the "message" list.
        """
        
        return self


    def GetSortImages(self):
        """
        Used by the ColumnSorterMixin, see wx/lib/mixins/listctrl.py.
        Not used by the "message" list.
        """

        # Returns the images used to display the sorting order
        return (self.sm_dn, self.sm_up)
    

    def HasManifest(self):
        """
        Returns whether the manifest file has been added to "other resources" or not.
        Used only for the "other_resources" list control. 
        """

        if self.GetItemCount() == 0:
            # No manifest file added, list is empty
            return False

        # If it exists, its text is in the 3rd column
        item = self.GetItem(0, 3)
        if item.GetText().find("manifest_template") >= 0:
            # Got it, is there
            return True

        # Shouldn't never reach this point, but who knows...
        return False
    

    def AddManifest(self):
        """
        Adds the manifest file to the list control.
        Used only for the "other_resources" list control. 
        """

        # The first item is always empty text with an informative icon
        indx = self.InsertImageStringItem(0, "", 0)
        # This is the format we need for the Setup.py file
        items = ["24", "1", "manifest_template%s"]
        for col, item in enumerate(items):
            self.SetStringItem(indx, col+1, item)

        # Update the data for the column sorter mixins
        self.SetItemData(indx, indx)
        self.itemDataMap[indx] = tuple(items)
        

    def DeleteManifest(self):
        """
        Deletes the manifest file from the list control.
        Used only for the "other_resources" list control. 
        """

        if not self.HasManifest():
            # How did we get here?!?
            return

        # Just as simple as it gets
        self.itemDataMap.pop(0)
        self.DeleteItem(0)


    def HandleNewResource(self, name):
        """ Handles the user request to add new resources. """

        if name.find("bitmap") >= 0:   # bitmap resources
            wildcard = "Bitmap files (*.bmp)|*.bmp"
        elif name.find("icon") >= 0:   # icon resources
            wildcard = "Icon files (*.ico)|*.ico"
        else:                          # whatever other resource
            wildcard = "All files (*.*)|*.*"

        # Run a file dialog with multiple selection            
        dlg = wx.FileDialog(self.MainFrame, message="New resources", wildcard=wildcard,
                            style=wx.FD_OPEN|wx.FD_MULTIPLE)

        # Show the dialog and retrieve the user response. If it is the OK response, 
        # process the data.
        if dlg.ShowModal() != wx.ID_OK:
            dlg.Destroy()
            return
        
        paths = dlg.GetPaths()
        dlg.Destroy()

        wx.BeginBusyCursor()

        # Get how many items we have in the current list control        
        numItems = self.GetItemCount()
        # Default value for the other_resources list control
        resourceType = 24
        # Get the number of columns in the current list control
        cols = self.GetColumnCount()

        config = []
        # loop over all the selected files
        for path in paths:
            values = []
            numItems += 1
            values.append(numItems)
            if cols == 4:
                # This is the other_resources list control
                values.append(resourceType)
            values.append(path)
            config.append(values)

        # Populate the list control
        self.PopulateList(config)
        # Update the project, something changed
        self.UpdateProject()

        wx.EndBusyCursor()
        

    def HandleNewModule(self):
        """ Handles the user request to add a new module. """

        # The first item is always empty text with an informative icon
        indx = self.InsertImageStringItem(sys.maxint, "", 0)
        # Insert an item with some visual indicator that should be edited...
        self.SetStringItem(indx, 1, "==> Edit Me! <==")
        # Update the column sorter mixin
        self.SetItemData(indx, indx)
        self.itemDataMap[indx] = ("", "==> Edit Me! <==")
        # Start editing the new label
        self.EditLabel(indx)
        

    def HandleDataFiles(self):
        """ Handles the user request to add new data files. """

        # Run a file dialog with multiple selection        
        wildcard = "All files (*.*)|*.*"
        dlg = wx.FileDialog(self.MainFrame, message="New data files", wildcard=wildcard,
                            style=wx.FD_OPEN|wx.FD_MULTIPLE)

        # Show the dialog and retrieve the user response. If it is the OK response, 
        # process the data.
        if dlg.ShowModal() != wx.ID_OK:
            dlg.Destroy()
            return
        
        paths = dlg.GetPaths()
        # Try to suggest a possible folder name to the user
        defaultDir = os.path.basename(dlg.GetDirectory())
        dlg.Destroy()

        # Ask the user which name the directory in the py2exe distribution
        # should have, suggesting the above defaultDir
        dlg = wx.TextEntryDialog(self.MainFrame, "Please enter the name of the folder for py2exe:",
                                 "Directory Name")
        dlg.SetValue(defaultDir)
        
        if dlg.ShowModal() != wx.ID_OK:
            # No choice made, no items added
            return

        # Get the user choice
        defaultDir = dlg.GetValue().strip()
        dlg.Destroy()

        if not defaultDir.strip():
            # Empty folder name?
            self.MainFrame.RunError("Error", "Invalid folder name.")
            return

        wx.BeginBusyCursor()
        config = [(defaultDir, paths)]

        # Populate the list control
        self.PopulateList(config)
        # Update the project, something changed
        self.UpdateProject()
        wx.EndBusyCursor()
        

    def GetSelectedIndices(self, state=wx.LIST_STATE_SELECTED):
        """ Returns the indices of the selected items in the list control. """

        indices = []
        lastFound = -1

        # Loop until told to stop        
        while 1:
            index = self.GetNextItem(lastFound, wx.LIST_NEXT_ALL, state)
            if index == -1:
                # No item selected
                break
            else:
                # Found one item, append to the list of condemned
                lastFound = index
                indices.append(index)

        return indices


    def TranslateToProject(self):
        """
        Translate all the list control item values to something understandable
        by the model stored in the database.
        """

        # Get all the information we need
        columns = self.GetColumnCount()
        windowName = self.GetName()
        values = []

        # Loop over all the items in the list control
        for row in xrange(self.GetItemCount()):
            temp = []
            # Loop over all the columns in the list control
            for col in xrange(1, columns):
                # Get the item text
                text = self.GetItem(row, col).GetText()
                if windowName.find("resources") >= 0 and col < columns-1:
                    # These are all integers, as py2exe requires
                    text = int(text)
                else:
                    text = text.encode("utf-8")
                temp.append(text)

            # depending on the list name, different output are requested            
            values.append((columns > 2 and [tuple(temp)] or [temp])[0])

        if columns == 2:
            # Transform the list of list in a flat 1D list of strings
            return flatten(values)

        if windowName == "data_files":
            # Throw away duplicated folder names
            return unique(values)
        
        return values


    def UpdateProject(self, changeIcon=True):
        """ Updates the project in the database, as something changed. """

        # Translate the list control values to something understandable
        # by the model stored in the database
        values = self.TranslateToProject()
        # Give feedback to the user if something changed (different icon)
        self.GetParent().GiveScreenFeedback(self.GetName(), values, changeIcon)
        

class CustomCodeViewer(wx.Frame):

    def __init__(self, parent, readOnly=False, text="", project=None, page=None):
        """
        Default class constructor.

        @param parent: the parent widget
        @param readOnly: indicates if the children StyledTextCtrl should be
                         read only or not
        @param text: the text to add to the StyledTextCtrl
        @param project: the project as it is stored in the database
        @param page: the page number of our main wx.aui.AuiNotebook

        """        

        wx.Frame.__init__(self, parent, title="GUI2Exe Code Viewer", size=(700, 550))

        # Store few objects, we are going to need them later        
        self.MainFrame = parent
        self.modified = False
        self.project = project
        self.page = page
        
        self.mainPanel = wx.Panel(self)
        # Add the StyledTextCtrl with Python syntax
        self.pythonStc = PythonSTC(self.mainPanel, readOnly)

        if not readOnly:
            # We are not read only, that means the user is going to add
            # some custom Python code in the py2exe Setup.py file
            saveBmp = self.MainFrame.CreateBitmap("code_save")
            discardBmp = self.MainFrame.CreateBitmap("code_discard")
            # Create a couple of themed buttons to save or discard changes
            self.saveButton = buttons.ThemedGenBitmapTextButton(self.mainPanel, -1, saveBmp, " Save ")
            self.discardButton = buttons.ThemedGenBitmapTextButton(self.mainPanel, -1, discardBmp, " Discard ")
            self.saveButton.SetDefault()

        # Do the hard work
        self.SetProperties(readOnly, text)
        self.LayoutItems(readOnly)
        self.BindEvents(readOnly)

        # Center ourselves in the main frame (GUI2Exe)
        self.CenterOnParent()
        self.Show()


    # ========================== #
    # Methods called in __init__ #
    # ========================== #
    
    def SetProperties(self, readOnly, text):
        """ Sets few properties for the frame and the StyledTextCtrl. """

        self.SetIcon(self.MainFrame.GetIcon())
        # Set the input text for the StyledTextCtrl
        self.pythonStc.SetText(text)
        self.pythonStc.EmptyUndoBuffer()
        # Colourise the Python code
        self.pythonStc.Colourise(0, -1)

        # line numbers in the margin
        self.pythonStc.SetMarginType(1, stc.STC_MARGIN_NUMBER)
        self.pythonStc.SetMarginWidth(1, 25)

        if readOnly:
            # No way, you can't change the text now
            self.pythonStc.SetReadOnly(True)
            

    def LayoutItems(self, readOnly):
        """ Layout the widgets with sizers. """
        
        mainSizer = wx.BoxSizer(wx.VERTICAL)
        panelSizer = wx.BoxSizer(wx.VERTICAL)
        stcSizer = wx.BoxSizer(wx.HORIZONTAL)
        rightSizer = wx.BoxSizer(wx.VERTICAL)

        if readOnly:
            topLabel = "This is the Setup file as it comes out after the pre-processing\n" \
                       "work done by GUI2Exe."
        else:
            topLabel = "Enter your custom code below and it will be inserted inside the\n" \
                       "Setup script. Note that you can use as 'keywords' also the compiler\n" \
                       "options like data_files, ignores and icon_resources."

        # The top label text is different depending on the choice of the user
        # If he/she chose to view the Setup.py file, it is read only
        topLabel = wx.StaticText(self.mainPanel, -1, topLabel)
        topLabel.SetFont(wx.Font(8, wx.DEFAULT, wx.NORMAL, wx.BOLD, 0, ""))
        panelSizer.Add(topLabel, 0, wx.ALL, 10)
        stcSizer.Add(self.pythonStc, 1, wx.ALL|wx.EXPAND, 5)

        if not readOnly:
            # Add the save/discard buttons
            rightSizer.Add((20, 20), 1, wx.EXPAND, 0)
            rightSizer.Add(self.saveButton, 0, wx.BOTTOM, 5)
            rightSizer.Add(self.discardButton, 0, wx.BOTTOM, 5)
            stcSizer.Add(rightSizer, 0, wx.ALL|wx.EXPAND, 5)

        # Layout everything
        panelSizer.Add(stcSizer, 1, wx.EXPAND, 0)
        self.mainPanel.SetSizer(panelSizer)
        panelSizer.Layout()
        # Add everything to the main sizer
        mainSizer.Add(self.mainPanel, 1, wx.EXPAND, 0)
        self.SetSizer(mainSizer)
        mainSizer.Layout()


    def BindEvents(self, readOnly):
        """ Binds the events for our CustomCodeViewer. """

        if not readOnly:
            self.Bind(wx.EVT_BUTTON, self.OnSave, self.saveButton)
            self.Bind(wx.EVT_BUTTON, self.OnDiscard, self.discardButton)
            self.Bind(wx.EVT_CLOSE, self.OnClose)
            # The user is going to add some custom code, so look for
            # possible modifications of it
            self.Bind(stc.EVT_STC_MODIFIED, self.OnModified, self.pythonStc)


    # ============== #
    # Event handlers #
    # ============== #
    
    def OnSave(self, event):
        """ Handles the wx.EVT_BUTTON event for CustomCodeViewer. """

        # Save the custom code added by the user in the project
        self.project.SetCustomCode(self.pythonStc.GetText())
        # Update the page itmap for our main wx.aui.AuiNotebook
        self.MainFrame.UpdatePageBitmap(self.project.GetName()+ "*", 1, self.page)
        # Change the modified flag, we are saved now
        self.modified = False
        
        event.Skip()


    def OnDiscard(self, event):
        """ Handles the wx.EVT_BUTTON event for CustomCodeViewer. """

        # No changes made or changes discarded. Destroy ourselves
        self.Destroy()
        event.Skip()


    def OnModified(self, event):
        """ Handles the stc.EVT_STC_MODIFIED event for CustomCodeViewer. """

        self.modified = True


    def OnClose(self, event):
        """ Handles the wx.EVT_CLOSE event for CustomCodeViewer. """

        if not self.modified:
            # No modifications or modifications saved. Destroy ourselves
            self.Destroy()
            return

        # Something changed. Ask for saving
        question = "The current code has changed.\nDo you want to save changes?"
        answer = self.MainFrame.RunError("Question", question)
        
        if answer == wx.ID_CANCEL:
            # Do you want to think about it, eh?
            return
        elif answer == wx.ID_YES:
            # Save the code in the project class
            self.project.SetCustomCode(self.pythonStc.GetText())
            self.MainFrame.SaveProject(self.project)

        event.Skip()



class PythonSTC(stc.StyledTextCtrl):

    def __init__(self, parent, readOnly):
        """
        Default class constructor.

        @param readOnly: indicates if the StyledTextCtrl should be in read-only mode

        """        

        stc.StyledTextCtrl.__init__(self, parent)

        # Retrieve the list control names: their names are set as keywords in
        # our StyledTextCtrl to highlight them to the user
        keys = _iconFromName.keys()[:]
        for key in _unWantedLists:
            # Remove the unwanted keys (such as "messages")
            keys.remove(key)

        # Add the Python keywords to the list        
        keys = keyword.kwlist + keys

        # Allow zooming (StyledTextCtrl is quite cool)        
        self.CmdKeyAssign(ord('B'), stc.STC_SCMOD_CTRL, stc.STC_CMD_ZOOMIN)
        self.CmdKeyAssign(ord('N'), stc.STC_SCMOD_CTRL, stc.STC_CMD_ZOOMOUT)

        # Set the Python lexer and the keywords
        self.SetLexer(stc.STC_LEX_PYTHON)
        self.SetKeyWords(0, " ".join(keys))

        self.SetProperty("fold", "1")
        # What the hell means this tab timmy???
        self.SetProperty("tab.timmy.whinge.level", "1")
        self.SetMargins(0,0)

        self.SetViewWhiteSpace(False)        
        self.SetEdgeMode(stc.STC_EDGE_BACKGROUND)
        if not readOnly:
            # Is the user enters custom code, it is good to remember him not
            # to exceed 80 characters per line :-D
            self.SetEdgeColumn(78)
        else:
            # We are seeing the automagically generated Setup.py file, which
            # can contain quite long strings...
            self.SetEdgeColumn(300)

        # Setup a margin to hold fold markers
        self.SetMarginType(2, stc.STC_MARGIN_SYMBOL)
        self.SetMarginMask(2, stc.STC_MASK_FOLDERS)
        self.SetMarginSensitive(2, True)
        self.SetMarginWidth(2, 12)

        # Like a flattened tree control using square headers
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPEN,    stc.STC_MARK_BOXMINUS,          "white", "#808080")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDER,        stc.STC_MARK_BOXPLUS,           "white", "#808080")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERSUB,     stc.STC_MARK_VLINE,             "white", "#808080")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERTAIL,    stc.STC_MARK_LCORNER,           "white", "#808080")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEREND,     stc.STC_MARK_BOXPLUSCONNECTED,  "white", "#808080")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPENMID, stc.STC_MARK_BOXMINUSCONNECTED, "white", "#808080")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERMIDTAIL, stc.STC_MARK_TCORNER,           "white", "#808080")

        self.Bind(stc.EVT_STC_UPDATEUI, self.OnUpdateUI)
        self.Bind(stc.EVT_STC_MARGINCLICK, self.OnMarginClick)
        
        # Make some styles,  The lexer defines what each style is used for, we
        # just have to define what each style looks like.  This set is adapted from
        # Scintilla sample property files.

        # Global default styles for all languages
        self.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "face:%(mono)s,size:%(size)d" % _faces)
        self.StyleClearAll()  # Reset all to be like the default

        # Global default styles for all languages
        self.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "face:%(helv)s,size:%(size)d" % _faces)
        self.StyleSetSpec(stc.STC_STYLE_LINENUMBER,  "back:#C0C0C0,face:%(helv)s,size:%(size2)d" % _faces)
        self.StyleSetSpec(stc.STC_STYLE_CONTROLCHAR, "face:%(other)s" % _faces)
        self.StyleSetSpec(stc.STC_STYLE_BRACELIGHT,  "fore:#FFFFFF,back:#0000FF,bold")
        self.StyleSetSpec(stc.STC_STYLE_BRACEBAD,    "fore:#000000,back:#FF0000,bold")

        # Python styles
        # Default 
        self.StyleSetSpec(stc.STC_P_DEFAULT, "fore:#000000,face:%(mono)s,size:%(size)d" % _faces)
        # Comments
        self.StyleSetSpec(stc.STC_P_COMMENTLINE, "fore:#007F00,face:%(mono)s,size:%(size)d" % _faces)
        # Number
        self.StyleSetSpec(stc.STC_P_NUMBER, "fore:#007F7F,size:%(size)d" % _faces)
        # String
        self.StyleSetSpec(stc.STC_P_STRING, "fore:#7F007F,face:%(mono)s,size:%(size)d" % _faces)
        # Single quoted string
        self.StyleSetSpec(stc.STC_P_CHARACTER, "fore:#7F007F,face:%(mono)s,size:%(size)d" % _faces)
        # Keyword
        self.StyleSetSpec(stc.STC_P_WORD, "fore:#00007F,bold,size:%(size)d" % _faces)
        # Triple quotes
        self.StyleSetSpec(stc.STC_P_TRIPLE, "fore:#7F0000,size:%(size)d" % _faces)
        # Triple double quotes
        self.StyleSetSpec(stc.STC_P_TRIPLEDOUBLE, "fore:#7F0000,size:%(size)d" % _faces)
        # Class name definition
        self.StyleSetSpec(stc.STC_P_CLASSNAME, "fore:#0000FF,bold,underline,size:%(size)d" % _faces)
        # Function or method name definition
        self.StyleSetSpec(stc.STC_P_DEFNAME, "fore:#007F7F,bold,size:%(size)d" % _faces)
        # Operators
        self.StyleSetSpec(stc.STC_P_OPERATOR, "bold,size:%(size)d" % _faces)
        # Identifiers
        self.StyleSetSpec(stc.STC_P_IDENTIFIER, "fore:#000000,face:%(mono)s,size:%(size)d" % _faces)
        # Comment-blocks
        self.StyleSetSpec(stc.STC_P_COMMENTBLOCK, "fore:#7F7F7F,size:%(size)d" % _faces)
        # End of line where string is not closed
        self.StyleSetSpec(stc.STC_P_STRINGEOL, "fore:#000000,face:%(mono)s,back:#E0C0E0,eol,size:%(size)d" % _faces)

        self.SetCaretForeground("BLUE")


    # ============== #
    # Event handlers #
    # ============== #
    
    def OnUpdateUI(self, evt):
        """ Handles the stc.EVT_STC_UPDATEUI event for PythonSTC. """
        
        # check for matching braces
        braceAtCaret = -1
        braceOpposite = -1
        charBefore = None
        caretPos = self.GetCurrentPos()

        if caretPos > 0:
            charBefore = self.GetCharAt(caretPos - 1)
            styleBefore = self.GetStyleAt(caretPos - 1)

        # check before
        if charBefore and chr(charBefore) in "[]{}()" and styleBefore == stc.STC_P_OPERATOR:
            braceAtCaret = caretPos - 1

        # check after
        if braceAtCaret < 0:
            charAfter = self.GetCharAt(caretPos)
            styleAfter = self.GetStyleAt(caretPos)

            if charAfter and chr(charAfter) in "[]{}()" and styleAfter == stc.STC_P_OPERATOR:
                braceAtCaret = caretPos

        if braceAtCaret >= 0:
            braceOpposite = self.BraceMatch(braceAtCaret)

        if braceAtCaret != -1  and braceOpposite == -1:
            self.BraceBadLight(braceAtCaret)
        else:
            self.BraceHighlight(braceAtCaret, braceOpposite)


    def OnMarginClick(self, evt):
        """ Handles the stc.EVT_STC_MARGINCLICK event for PythonSTC. """
        
        # fold and unfold as needed
        if evt.GetMargin() == 2:
            if evt.GetShift() and evt.GetControl():
                self.FoldAll()
            else:
                lineClicked = self.LineFromPosition(evt.GetPosition())

                if self.GetFoldLevel(lineClicked) & stc.STC_FOLDLEVELHEADERFLAG:
                    if evt.GetShift():
                        self.SetFoldExpanded(lineClicked, True)
                        self.Expand(lineClicked, True, True, 1)
                    elif evt.GetControl():
                        if self.GetFoldExpanded(lineClicked):
                            self.SetFoldExpanded(lineClicked, False)
                            self.Expand(lineClicked, False, True, 0)
                        else:
                            self.SetFoldExpanded(lineClicked, True)
                            self.Expand(lineClicked, True, True, 100)
                    else:
                        self.ToggleFold(lineClicked)


    # ================= #
    # Auxiliary methods #
    # ================= #
    
    def FoldAll(self):
        """ Folds/unfolds everything in the code. """
        
        lineCount = self.GetLineCount()
        expanding = True

        # find out if we are folding or unfolding
        for lineNum in range(lineCount):
            if self.GetFoldLevel(lineNum) & stc.STC_FOLDLEVELHEADERFLAG:
                expanding = not self.GetFoldExpanded(lineNum)
                break

        lineNum = 0

        while lineNum < lineCount:
            level = self.GetFoldLevel(lineNum)
            if level & stc.STC_FOLDLEVELHEADERFLAG and \
               (level & stc.STC_FOLDLEVELNUMBERMASK) == stc.STC_FOLDLEVELBASE:

                if expanding:
                    self.SetFoldExpanded(lineNum, True)
                    lineNum = self.Expand(lineNum, True)
                    lineNum = lineNum - 1
                else:
                    lastChild = self.GetLastChild(lineNum, -1)
                    self.SetFoldExpanded(lineNum, False)

                    if lastChild > lineNum:
                        self.HideLines(lineNum+1, lastChild)

            lineNum = lineNum + 1



    def Expand(self, line, doExpand, force=False, visLevels=0, level=-1):
        """ Expands/contracts lines after a folding event. """
        
        lastChild = self.GetLastChild(line, level)
        line = line + 1

        while line <= lastChild:
            if force:
                if visLevels > 0:
                    self.ShowLines(line, line)
                else:
                    self.HideLines(line, line)
            else:
                if doExpand:
                    self.ShowLines(line, line)

            if level == -1:
                level = self.GetFoldLevel(line)

            if level & stc.STC_FOLDLEVELHEADERFLAG:
                if force:
                    if visLevels > 1:
                        self.SetFoldExpanded(line, True)
                    else:
                        self.SetFoldExpanded(line, False)

                    line = self.Expand(line, doExpand, force, visLevels-1)

                else:
                    if doExpand and self.GetFoldExpanded(line):
                        line = self.Expand(line, True, force, visLevels-1)
                    else:
                        line = self.Expand(line, False, force, visLevels-1)
            else:
                line = line + 1

        return line

        
class Py2ExeMissing(wx.Frame):

    def __init__(self, parent, project, dll=False):
        """
        Default class constructor.

        @param parent: the parent widget
        @param project: the project as stored in the database
        @param dll: indicates whether we are going to show the missing binary
                    dependencies (dll) or the missing modules (py)

        """
        
        wx.Frame.__init__(self, parent, size=(500, 400))
        
        self.MainFrame = parent
        self.mainPanel = wx.Panel(self)

        # Build column names dinamically depending on the dll value
        name = (dll and ["binarydependencies"] or ["missingmodules"])[0]
        if dll:
            columnNames = ["DLL Name", "DLL Path"]
        else:
            # There might be some strange sub-sub-sub module imported...
            columnNames = ["Main Module", "Sub-Module 1", "Sub-Module 2", "Sub-Module 3"]

        # Build the base list control            
        self.list = BaseListCtrl(self.mainPanel, columnNames, name, self.MainFrame)

        # Build a couple of fancy and useless buttons        
        okBmp = self.MainFrame.CreateBitmap("project_ok")
        cancelBmp = self.MainFrame.CreateBitmap("exit")
        self.okButton = buttons.ThemedGenBitmapTextButton(self.mainPanel, -1, okBmp, " Ok ")
        self.cancelButton = buttons.ThemedGenBitmapTextButton(self.mainPanel, -1, cancelBmp, " Cancel ")

        # Do the hard work        
        self.SetProperties(dll)
        self.LayoutItems(dll)
        self.BindEvents()

        # Populate the list, retrieving the data directly from the project
        missingModules, binaryDependencies = project.GetCompilationData()
        self.list.PopulateList((dll and [binaryDependencies] or [missingModules])[0])

        for col in xrange(2, self.list.GetColumnCount()):
            self.list.SetColumnWidth(col, wx.LIST_AUTOSIZE)
            
        self.CenterOnParent()
        self.Show()
        

    # ========================== #
    # Methods called in __init__ #
    # ========================== #
    
    def SetProperties(self, dll):
        """ Sets few properties for the Py2ExeMissing frame. """        

        if dll:
            # We are showing binary dependencies
            title = "Py2Exe Binary Dependencies"
        else:
            # These are what py2exe thinks are the missing modules
            title = "Py2Exe Missing Modules"
            
        self.SetTitle(title)
        self.SetIcon(self.MainFrame.GetIcon())

        # We want the second column to fill all the available space
        self.list.setResizeColumn(2)    
        self.okButton.SetDefault()


    def LayoutItems(self, dll):
        """ Layouts the widgets with sizers. """        

        mainSizer = wx.BoxSizer(wx.VERTICAL)
        panelSizer = wx.BoxSizer(wx.VERTICAL)
        bottomSizer = wx.BoxSizer(wx.HORIZONTAL)

        if dll:
            # We are showing binary dependencies
            label = "Make sure you have the license if you distribute any of them, and\n"  \
                    "make sure you don't distribute files belonging to the operating system."
        else:
            # These are what py2exe thinks are the missing modules
            label = "Py2Exe thinks that these modules (and sub-modules) are missing.\n" \
                    "Inclusion of one or more of them may allow your compiled application to run."

        label = wx.StaticText(self.mainPanel, -1, label)
        label.SetFont(wx.Font(8, wx.DEFAULT, wx.NORMAL, wx.BOLD, 0, ""))

        panelSizer.Add(label, 0, wx.ALL, 10)
        panelSizer.Add(self.list, 1, wx.EXPAND|wx.ALL, 5)

        # Add the fancy and useless buttons
        bottomSizer.Add(self.okButton, 0, wx.ALL, 15)
        bottomSizer.Add((0, 0), 1, wx.EXPAND)
        bottomSizer.Add(self.cancelButton, 0, wx.ALL, 15)

        panelSizer.Add(bottomSizer, 0, wx.EXPAND)
        self.mainPanel.SetSizer(panelSizer)
        panelSizer.Layout()

        # Add everything to the main sizer        
        mainSizer.Add(self.mainPanel, 1, wx.EXPAND)
        self.SetSizer(mainSizer)
        mainSizer.Layout()


    def BindEvents(self):
        """ Binds the events for Py2ExeMissing. """

        self.Bind(wx.EVT_BUTTON, self.OnOk, self.okButton)
        self.Bind(wx.EVT_BUTTON, self.OnOk, self.cancelButton)
        

    # ============== #
    # Event handlers #
    # ============== #
    
    def OnOk(self, event):
        """ Handles the wx.EVT_BUTTON event for Py2ExeMissing. """

        # Very useful these buttons, eh?
        self.Destroy()
        event.Skip()

